const {Tuple2} = require('fantasy-tuples')
const Either = require('fantasy-eithers')
const {tagged} = require('daggy')

const Parser = tagged('innerFn', 'label')

const pchar = function pchar (ch) {
    let innerFn = function (input) {
        if (input === undefined || input === "")
            return Either.Left(Tuple2(ch, "no more input"))
        else {
            let first = input[0]
            if (first === ch){
                let remaining = input.substring(1)
                return Either.Right(Tuple2(ch, remaining))
            } else {
                return Either.Left(Tuple2(ch, `Unexpected  '${first}'.`))
            }
        }
    }
    return Parser(innerFn, ch)
}

// const satisfy = function satisfy (predicate) {
//     let innerFn = function (input) {
//         if (input === undefined || input === "")
//             return Either.Left("no more input")
//         else {
//             let first = input[0]
//             if (first === ch){
//                 let remaining = input.substring(1)
//                 return Either.Right(Tuple2(ch, remaining))
//             } else {
//                 return Either.Left(Tuple2(ch, `Expecting ${ch}. Got ${first}.`))
//             }
//         }
            
//     }
//     return Parser(innerFn, "")
// }


const run = function run (parser, input) {
    return parser.innerFn(input)
}

const andThen = function andThen(p1, p2) {
    let innerFn = function(input) {
        let result1 = run(p1, input)
        if(result1.l) {
            return Either.Left(result1.l)
        } else {
            let remaining1 = result1.r._2
            let result2 = run (p2, remaining1)
            if (result2.l) {
                return Either.Left(result2.l)
            } else {
                let combinedVal = Tuple2(result1.r._1, result2.r._1)
                return Either.Right(Tuple2(combinedVal, result2.r._2))
            }
        }
    }
    
    return Parser(innerFn, `${p1.label} andThen ${p2.label}`)
}

const orElse = function orElse(p1, p2) {
    let innerFn = function(input) {
        let result1 = run(p1, input)
        if(result1.r) {
            return Either.Right(result1.r)
        } else {
            let result2 = run (p2, input)
            if (result2.l) {
                return Either.Left(result2.l)
            } else {
                return Either.Right(result2.r)
            }
        }
    }
    return Parser(innerFn, `${p1.label} orElse ${p2.label}`)
}

const mapP = function mapP (fn, parser) {
    let innerFn = function(input) {
        let result = run(parser, input)
        if(result.r) {
            let newValue = fn(result.r._1)
            return Either.Right(Tuple2(newValue, result.r._2))
        } else {
            return Either.Left(result.l)
        }
    }
    return Parser(innerFn, `map over ${parser.label}`)    
}

const choice = function(listOfParsers) {
    let labels = listOfParsers.map(p => p.label)
    return setLabel( listOfParsers.reduce(orElse) , `choice of [${labels}]` )
}

const anyOf = function(listOfChars) {
    return setLabel( choice(listOfChars.map(pchar)), `any of [${listOfChars}]` )
}


const sequence = function(listOfParsers) {
    let labels = listOfParsers.map(p => p.label)
    let concatResults = function(p1, p2){
        let res = andThen(p1, p2)
        return mapP(lists => lists._1.concat(lists._2), res) 
    } 

    let resultingParser = listOfParsers.map(p => mapP(r => [r], p))
    .reduce(concatResults)
    return setLabel(resultingParser, `sequence of [${labels}]`)
}

const pstring = function(str) {
    let strp = str.split('')
    .map(pchar)
    let resultingParser = mapP((lst => lst.join('')), sequence(strp))
    return setLabel(resultingParser,    `string "${str}"`)
}

const parseZeroOrMore = function(parser, input) {
    let firstResult = run(parser, input)
    if(firstResult.l) {
         return Tuple2([], input)
    } else {
        let firstValue = firstResult.r._1
        let inputAfterFirstParse = firstResult.r._2
        let subsequent = parseZeroOrMore(parser, inputAfterFirstParse)
        let values = [firstValue].concat(subsequent._1)
        return Tuple2(values, subsequent._2)
    }
}

const many = function (parser) {
    let innerFn = function (input) {
        return Either.Right(parseZeroOrMore(parser, input))
    }
    return Parser(innerFn, `zero or more ${parser.label}`)
}

const many1 = function (parser) {
    let innerFn = function (input) {
        let firstResult = run(parser, input)
        if (firstResult.l) {
            return Either.Left(firstResult.l)
        } else {
            let firstValue = firstResult.r._1
            let inputAfterFirstParse = firstResult.r._2

            let subsequent = parseZeroOrMore(parser, inputAfterFirstParse)
            let values = [firstValue].concat(subsequent._1)
            return Either.Right(Tuple2(values, subsequent._2))
        }
    }
    return Parser(innerFn, `one or more ${parser.label}`)
}

const opt = function(parser) {
    let innerFn = function (input) {
        let firstResult = run(parser, input)
        if (firstResult.l) {
            return Either.Rigth(Tuple2([], input))
        } else {
            return Either.Right(Tuple2([firstResult.r._1], firstResult.r._2))
        }
    }
    return Parser(innerFn, `zero or one ${parser.label}`)
}

const throwLeft = function(p1, p2) {
    return setLabel( mapP (t => t._2, andThen(p1, p2)), `${p2.label} preceded by ${p1.label}` ) 
}

const throwRight = function(p1, p2) {
    return setLabel( mapP (t => t._1, andThen(p1, p2)), `${p1.label} followed by ${p2.label}` )
}

const between = function(p1, p2, p3) {
    return setLabel( throwRight(throwLeft(p1, p2), p3), `${p2.label} between ${p1.label} and ${p3.label}` )
}

const setLabel = function(parser, newLabel) {
    let newInnerFn = function(input) {
        let result = parser.innerFn(input)
        if (result.r) {
            return Either.Right(result.r)
        } else {
            return Either.Left(Tuple2(newLabel, result.l._2))
        }
    }
    return Parser(newInnerFn, newLabel)
}



module.exports = {Parser, run, pchar, andThen, orElse, mapP, choice, 
    anyOf, sequence, pstring, parseZeroOrMore, many, many1, opt, 
    throwLeft, throwRight, between, setLabel}